from .models import ChessBert, CharBert, BertPool
from ._loss import loss
from .decode import argmax_ucis, greedy_pgn
from . import chars, words

__all__ = [
  'ChessBert', 'CharBert', 'BertPool',
  'loss',
  'argmax_ucis', 'greedy_pgn',
  'chars', 'words',
]


from .vocab import legal, remove_symbols, SpecialToken, Vocabulary, legal_styled
from .samples import parse_sample, Sample

__all__ = [
  'legal', 'remove_symbols', 'SpecialToken', 'Vocabulary',
  'parse_sample', 'Sample', 'legal_styled',
]


from typing import Sequence, NamedTuple, Literal
import chess_utils as cu
from .._labels import Label, labels

class TokenizedInput(NamedTuple):
	tokens: Sequence[str | Literal['[SEP]']]
	ends: Sequence[int]

def tokenize(words: Sequence[str]) -> TokenizedInput:
	"""Split words into characters, adding '[SEP]' between words. `return.ends` are the original word end indices"""
	tokens = []
	ends = []
	end = 0
	for word in words:
		tokens.extend(word)
		tokens.append('[SEP]')
		end += len(word)
		ends.append(end)
	
	return TokenizedInput(tokens, ends)
	
class Sample(NamedTuple):
	inputs: TokenizedInput
	labs: Sequence[Label]

def parse_sample(inputs: str, ucis: str | None = None) -> Sample:
	sans = inputs.strip('\n').replace('+', '').replace('#', '').split(' ')
	toks = tokenize(sans)
	parsed_ucis = cu.sans2ucis(sans) if ucis is None else ucis.strip('\n').split(' ')
	labs = labels(parsed_ucis)
	return Sample(toks, labs)


from typing import Iterable, Sequence
import fs

def iterate_samples(
  input_files: Sequence[str], uci_files: Sequence[str],
) -> Iterable[tuple[str, str]]:
  inputs = fs.concat_lines(input_files)
  ucis = fs.concat_lines(uci_files)
  yield from zip(inputs, ucis)



from typing import Literal, Mapping, TypeAlias, Iterable, Sequence
import chess_notation as cn
from .sans import legal_sans, sort_key

SpecialToken: TypeAlias = Literal['[PAD]', '[CLS]', '[SEP]', '[MASK]']
PAD: TypeAlias = Literal['[PAD]']
MASK: TypeAlias = Literal['[MASK]']
Vocabulary: TypeAlias = Mapping[str | SpecialToken, int]

def remove_symbols(san: str) -> str:
  """Remove check and mate symbols from a SAN move"""
  return san.replace('+', '').replace('#', '')

def make_vocab(words: Iterable[str]) -> Vocabulary:
  return { word: i for i, word in enumerate(words) }

def legal(with_symbols: bool = False, *, pad_token: str = '[PAD]') -> Vocabulary:
  """Vocabulary containing all legal SAN moves
  - `with_symbols`: whether to include `+` and `#` (triples the size of the vocabulary, though)
  """
  return make_vocab(legal_sans(with_symbols) + [pad_token])

def legal_styled(
  with_symbols: bool = False, *,
  motions: cn.MotionStyles = cn.MotionStyles(),
  effects: cn.KingEffectStyles = cn.KingEffectStyles(),
  languages: Sequence[cn.Language] = cn.LANGUAGES,
  pad_token: str = '[PAD]',
  extra_tokens: Sequence[str] = ['[UNK]']
) -> Vocabulary:
  """All legal sans, styled. Size 60953 with default parameters
  - Legal SANs come first, then all representations
  """
  legal = legal_sans(with_symbols)
  words = set()
  for san in legal:
    words |= cn.all_representations(san, motions=motions, effects=effects, languages=languages)
  new_words = words - set(legal)
  sorted_words = sorted(new_words, key=sort_key)
  return make_vocab(legal + [pad_token] + sorted_words + list(extra_tokens))
  


from .pos_enc import positional_encoding
from .segmented import segment_encoding, pool_batch, pool_sample
from .data import Batch, Sample, sample_from, collate_same, collate_separate

__all__ = [
  'positional_encoding', 'segment_encoding',
  'pool_batch', 'pool_sample',
  'sample_from', 'Batch', 'Sample', 'collate_same', 'collate_separate'
]


from dataclasses import dataclass
import os
import json

@dataclass
class Experiment:
  path: str
  params: dict
  replace: bool = False

  def __post_init__(self):
    os.makedirs(self.path, exist_ok=self.replace)
    mode = 'w' if self.replace else 'x'
    with open(f'{self.path}/params.json', mode) as f:
      json.dump(self.params, f)

    with open(f'{self.path}/logs.csv', mode) as f:
      f.write('step,loss\n')

  def log(self, step: int, loss: float):
    with open(f'{self.path}/logs.csv', 'a') as f:
      f.write(f'{step},{loss}\n')


from .metrics import accuracy, uci_accuracy
from ._labels import labels, Label
from .decode import decode_ucis
from .experiment import Experiment
from . import pytorch, dataset, chars, words

__all__ = [
  'accuracy', 'uci_accuracy', 'labels', 'Label', 'decode_ucis',
  'pytorch', 'dataset', 'chars', 'words',
  'Experiment',
]


from typing import Sequence
import os
from glob import glob
from pydantic import BaseModel

class Meta(BaseModel):
  class Files(BaseModel):
    san: str | None = None
    uci: str
    styled: str | None = None
  samples: int
  source: str
  files: Files

def read_meta(base: str) -> Meta:
  with open(os.path.join(base, 'meta.json')) as f:
    return Meta.model_validate_json(f.read())
  
def input_files(base: str, meta: Meta | None = None) -> Sequence[str]:
  meta = meta or read_meta(base)
  files = meta.files.styled or meta.files.san
  if files is None:
    raise ValueError('No input files found in meta')
  return glob(os.path.join(base, files))

def uci_files(base: str, meta: Meta | None = None) -> Sequence[str]:
  meta = meta or read_meta(base)
  return glob(os.path.join(base, meta.files.uci))
